@@ -162,9 +162,11 @@ function send() { // periodically send events and update_canvas and request obje
 }
 
 // *************************************************************************************************** //
-// THE REST OF THIS FILE IS IDENTICAL IN glowcomm.html AND glowcomm.js (except for the last few lines) //
+// THE REST OF THIS FILE IS NEARLY IDENTICAL IN glowcomm.html AND glowcomm.js AND glowcommlab.js //
 
-// Should eventually have glowcomm.html and glowcom.js both import this common component.
+// Should eventually have glowcomm.html, glowcom.js, and glowcommlab.js all import this common component.
+
+window.__GSlang = "vpython"
 
 function msclock() {
     "use strict";
@@ -186,7 +188,7 @@ var lastrange = 1
 var lastautoscale = true
 var lastsliders = {}
 var lastkeysdown = []
-var interval = 33 // milliseconds
+var interval = 17 // milliseconds
 
 function update_canvas() { // mouse location and other stuff
     "use strict";
@@ -369,7 +371,7 @@ function process_waitfor(event) {
 }
 
 function process_binding(event) { // event associated with a previous bind command
-    "use strict";
+	"use strict";
     event.bind = true
 	process(event)
 }
@@ -410,7 +412,7 @@ function control_handler(obj) {  // button, menu, slider, radio, checkbox, winpu
 }
 
 // attrs are X in {'a': '23X....'} avaiable: none
-var attrs = {'a':'pos', 'b':'up', 'c':'color', 'd':'trail_color', // don't use single and double quotes; available: comma, but maybe that would cause trouble
+let attrs = {'a':'pos', 'b':'up', 'c':'color', 'd':'trail_color', // don't use single and double quotes; available: comma, but maybe that would cause trouble
          'e':'ambient', 'f':'axis', 'g':'size', 'h':'origin', 'i':'textcolor',
          'j':'direction', 'k':'linecolor', 'l':'bumpaxis', 'm':'dot_color',
          'n':'foreground', 'o':'background', 'p':'ray', 'E':'center', '#':'forward', '+':'resizable',
@@ -437,34 +439,35 @@ var attrs = {'a':'pos', 'b':'up', 'c':'color', 'd':'trail_color', // don't use s
          '?':'font', '/':'texture'}
          
 // attrsb are X in {'b': '23X....'}; ran out of easily typable one-character codes
-var attrsb = {'a':'userzoom', 'b':'userspin', 'c':'range', 'd':'autoscale', 'e':'fov',
+let attrsb = {'a':'userzoom', 'b':'userspin', 'c':'range', 'd':'autoscale', 'e':'fov',
               'f':'normal', 'g':'data', 'h':'checked', 'i':'disabled', 'j':'selected',
               'k':'vertical', 'l':'min', 'm':'max', 'n':'step', 'o':'value',
               'p':'left', 'q':'right', 'r':'top', 's':'bottom', 't':'_cloneid',
               'u':'logx', 'v':'logy', 'w':'dot', 'x':'dot_radius', 
               'y':'markers', 'z':'legend', 'A':'label','B':'delta', 'C':'marker_color',
-              'D':'size_units', 'E':'userpan', 'F':'scroll', 'G':integrate_selected}
+              'D':'size_units', 'E':'userpan', 'F':'scroll', 'G':'choices', 'H':'depth',
+			  'I':'round', 'J':'name', 'K':'offset', 'L':'attach_idx'}
 
-// methods are X in {'m': '23X....'}
-var methods = {'a':'select', 'b':'pos', 'c':'start', 'd':'stop', 'f':'clear', // unused eghijklmnopvxyzCDFAB
+// methods are X in {'m': '23X....'} available: u
+let methods = {'a':'select', 'b':'pos', 'c':'start', 'd':'stop', 'f':'clear', // unused eghijklmnopvxyzCDFAB
 			   'q':'plot', 's':'add_to_trail',
-               't':'follow', 'u':'_attach_arrow', 'w':'clear_trail',
+               't':'follow', 'w':'clear_trail',
                'G':'bind', 'H':'unbind', 'I':'waitfor', 'J':'pause', 'K':'pick', 'L':'GSprint',
 		       'M':'delete', 'N':'capture'}
          
-var vecattrs = ['pos', 'up', 'color', 'trail_color', 'axis', 'size', 'origin', '_attach_arrow',
+let vecattrs = ['pos', 'up', 'color', 'trail_color', 'axis', 'size', 'origin', 
                 'direction', 'linecolor', 'bumpaxis', 'dot_color', 'ambient', 'add_to_trail', 'textcolor',
                 'foreground', 'background', 'ray', 'ambient', 'center', 'forward', 'normal',
-                'marker_color']
+                'marker_color', 'offset']
                 
-var textattrs = ['text', 'align', 'caption', 'title', 'title_align', 'xtitle', 'ytitle', 'selected', 'capture',
-                 'label', 'append_to_caption', 'append_to_title', 'bind', 'unbind', 'pause', 'GSprint']
+let textattrs = ['text', 'align', 'caption', 'title', 'title_align', 'xtitle', 'ytitle', 'selected', 'capture', 'name',
+                 'label', 'append_to_caption', 'append_to_title', 'bind', 'unbind', 'pause', 'GSprint', 'choices']
 
 // patt gets idx and attr code; vpatt gets x,y,z of a vector            
-var patt = /(\d+)(.)(.*)/
-var vpatt = /([^,]*),([^,]*),(.*)/
-var quadpatt = /([^,]*),([^,]*),(.*)/
-var plotpatt = /([^,]*),([^,]*)/
+const patt = /(\d+)(.)(.*)/
+const vpatt = /([^,]*),([^,]*),(.*)/
+const quadpatt = /([^,]*),([^,]*),(.*)/
+const plotpatt = /([^,]*),([^,]*)/
 
 function decode(data) { 
     "use strict";
@@ -476,12 +479,12 @@ function decode(data) {
 	var ms = []
 	
 	if ('attrs' in data) {
-		var c = data['attrs']
+		let c = data['attrs']
 		for (i=0; i<c.length; i++) { // step through the encoded attributes and methods
-			var d = c[i]
+			let d = c[i]
 			// constructor or appendcmd not currently compressed
-			var whichlist = d[0] // 'a' or 'b' or 'm'
-			var datatype = (whichlist == 'm') ? 'method' : 'attr'
+			let whichlist = d[0] // 'a' or 'b' or 'm'
+			let datatype = (whichlist == 'm') ? 'method' : 'attr'
 			s = d.slice(1)
 			m = s.match(patt)
 			idx = Number(m[1])
@@ -502,8 +505,16 @@ function decode(data) {
                     vs = [Number(val[1]), Number(val[2]), Number(val[3]), Number(val[4])]
                 }
 			} else if (textattrs.indexOf(attr) > -1) {
-                // '\n' doesn't survive JSON transmission, so in vpython.py we replace '\n' with '<br>'
-				val = m[3].replace(/<br>/g, "\n")
+                if (attr == 'choices') { // menu choices are wrapped in a list
+                    val = m[3].slice(2,-2)
+                    val = val.replace(/'/g, '') // remove quotes
+                    let s = val.split(',')
+                    val = []
+                    for (let a of s) {val.push(a)}
+                } else {
+                    // '\n' doesn't survive JSON transmission, so in vpython.py we replace '\n' with '<br>'
+                    val = m[3].replace(/<br>/g, "\n")
+                }
 			} else if (attr == 'rotate') { // angle,x,y,z,x,y,z
 				var temp = m[3]
 				val = []
@@ -526,6 +537,9 @@ function decode(data) {
 				}
 			} else if (attr == 'waitfor' || attr == 'pause' || attr == 'delete') {
 				val = m[3]
+            } else if (attr == 'follow') {
+                if (m[3] == 'None') val = null
+                else val = Number(m[3])
 			} else val = Number(m[3])
 			out = {'idx':idx, 'attr':attr, 'val':val}
 			if (datatype == 'attr') as.push(out)
@@ -575,7 +589,9 @@ function handler(data) {
 	for (var d in data) {
 		for (var i in data[d]) console.log(i, JSON.stringify(data[d][i]))
 	}
-    */
+	*/
+	
+	
 	
 	if (data.cmds !== undefined && data.cmds.length > 0) handle_cmds(data.cmds)
 	if (data.methods !== undefined && data.methods.length > 0) handle_methods(data.methods)
@@ -602,7 +618,8 @@ function handle_cmds(dcmds) {
 		//assembling cfg
 		var vlst = ['pos', 'color', 'size', 'axis', 'up', 'direction', 'center', 'forward', 'foreground',
 				 'background', 'ambient', 'linecolor', 'dot_color', 'trail_color', 'textcolor', 'attrval',
-				 'origin', 'normal', 'bumpaxis','texpos', 'start_face_color', 'end_face_color', 'marker_color']
+				 'origin', 'normal', 'bumpaxis','texpos', 'start_face_color', 'end_face_color', 'marker_color',
+				 'start_normal', 'end_normal', 'offset']
 		if ((obj != 'gcurve') && ( obj != 'gdots' ) ) vlst.push( 'size' )
 		var cfg = {}
 		var objects = []
@@ -616,14 +633,16 @@ function handle_cmds(dcmds) {
 				   cfg[attr] = o2vec3(val)
 				}                            
 			} else if ( (attr == 'pos' && (obj == 'curve' || obj == 'points')) ||
-						(attr == 'path' && obj == 'extrusion') ) { // only occurs in constructor
-				var ptlist = []
-				for (var kk = 0; kk < val.length; kk++) {
-					ptlist.push( o2vec3(val[kk]) )
+						(obj == 'extrusion' && (attr == 'path' || attr == 'color') ) ) { // only occurs in constructor
+				let ptlist = []
+				if (val[0].length === undefined) { // a single triple [x,y,z] to convert to a vector
+					ptlist = o2vec3(val)
+				} else {
+					for (var kk = 0; kk < val.length; kk++) {
+						ptlist.push( o2vec3(val[kk]) )
+					}
 				}
 				cfg[attr] = ptlist
-			} else if (attr === "axis" && obj == 'arrow') {
-				cfg['axis_and_length'] = o2vec3(val) 
 			} else if (vlst.indexOf(attr) !== -1) {
 				cfg[attr] = o2vec3(val)
 			} else if (triangle_quad.indexOf(attr) !== -1) {
@@ -679,6 +698,8 @@ function handle_cmds(dcmds) {
 		}
 		// creating the objects
 		cfg.idx = idx // reinsert idx, having looped thru all other attributes
+		// triangle and quad objects should not have a canvas attribute; canvas is provided in the vertex objectsE
+		if ((obj == 'triangle' || obj == 'quad') && cfg.canvas !== undefined) delete cfg.canvas
 		switch (obj) {
 			case 'box':           {glowObjs[idx] = box(cfg); break}
 			case 'sphere':        {glowObjs[idx] = sphere(cfg); break}
@@ -752,41 +773,40 @@ function handle_cmds(dcmds) {
 				}
 				break
 			}
-			case 'local_light':   {glowObjs[idx] = local_light(cfg); break}
+			case 'local_light': {
+				let g = glowObjs[idx] = local_light(cfg)
+				if (cfg.offset !== undefined) { // mocking up attach_light
+					g.__obj = glowObjs[cfg.attach_idx]
+					g.canvas.attached_lights.push(g)
+				}
+				break
+			}
 			case 'distant_light': {glowObjs[idx] = distant_light(cfg); break}
 			case 'canvas':        {
-				var container = document.getElementById("glowscript");
-				if (container !== null) {
-					window.__context = { glowscript_container: $("#glowscript").removeAttr("id")}
-				}
-				glowObjs[idx] = canvas(cfg)
-				glowObjs[idx]['idx'] = idx
-				try{
-					glowObjs[idx].wrapper[0].addEventListener("contextmenu", function(event){
-						event.preventDefault(); 
-						event.stopPropagation(); 
-					});
-				}
-				catch(err) {
-					console.log("glowcomm canvas contextmenu event : ",err.message);
+				if ((typeof isjupyterlab_vpython !== 'undefined') && (isjupyterlab_vpython === true)) {
+					var container = document.getElementById("glowscript");
+					if (container !== null) {
+						window.__context = { glowscript_container: $("#glowscript").removeAttr("id")}
+					}
+					glowObjs[idx] = canvas(cfg)
+					glowObjs[idx]['idx'] = idx
+					try{
+						glowObjs[idx].wrapper[0].addEventListener("contextmenu", function(event){
+							event.preventDefault(); 
+							event.stopPropagation(); 
+						});
+					}
+					catch(err) {
+						console.log("glowcomm canvas contextmenu event : ",err.message);
+					}
+				} else {
+					glowObjs[idx] = canvas(cfg)
+					glowObjs[idx]['idx'] = idx
 				}
 				break
 					// Display frames per second and render time:
 					//$("<div id='fps'/>").appendTo(glowObjs[idx].title)
 			}
-			case 'attach_arrow':  {
-                var attrs = ['pos', 'size', 'axis', 'up', 'color']
-				var o = glowObjs[cfg['obj']]
-				delete cfg['obj']
-				var attr = cfg['attr']
-				delete cfg['attr']
-                var val = cfg['attrval']
-                delete cfg['attrval']
-                if (attrs.indexOf(attr) < 0) attr = '_attach_arrow'
-                o.attr = val
-				glowObjs[idx] = attach_arrow( o, attr, cfg )
-				break
-			}
 			case 'attach_trail': {
 				if ( typeof cfg['_obj'] === 'string' ) {
 					var o = cfg['_obj'] // the string '_func'
@@ -818,10 +838,13 @@ function handle_cmds(dcmds) {
 				break
 			}
 			case 'radio': {
+				cfg.canvas = canvas.get_selected()
+				cfg = fix_location(cfg)
+				delete cfg.canvas
 				cfg.objName = obj
 				cfg.bind = control_handler
-				cfg = fix_location(cfg)
 				glowObjs[idx] = radio(cfg)
+				// glowObjs[idx].canvas = canvas.get_selected()
 				break
 			}
 			case 'button': {
@@ -851,60 +874,18 @@ function handle_cmds(dcmds) {
 			default:
 				console.log("Unable to create object")
 		}
-		
-		/*
-		if (obj === 'redisplay') {
-			var c = document.getElementById(cmd.sceneId)
-			if (c !== null) {
-				var scn = "#" + cmd.sceneId
-				glowObjs[idx].sceneclone = $(scn).clone(true,true)
-				//document.getElementById('glowscript2').appendChild(c)
-				//document.getElementById('glowscript2').replaceWith(c)
-				$('#glowscript2').replaceWith(c)
-				c = document.getElementById(cmd.sceneId)
-				var cont = scn + " .glowscript"
-				window.__context = { glowscript_container:    $(cont) }
-			} else {
-				window.__context = { glowscript_container: $("#glowscript").removeAttr("id") }                    
-				var newcnvs = canvas()
-				for (var obj in glowObjs[idx].objects) {
-					var o = glowObjs[idx].objects[obj]
-					if ((o.constructor.name !== 'curve') && (o.constructor.name !== 'points')) {
-						glowObjs[o.gidx] = o.clone({canvas: newcnvs})
-						var olen = newcnvs.objects.length
-						if (olen > 0) {
-							newcnvs.objects[olen - 1].gidx = o.gidx
-						}
-					}
-				}
-				glowObjs[idx] = newcnvs
-				$("#glowscript2").attr("id",cmd.sceneId)
-			}
-		} else if (obj === 'delete') {
-			b = glowObjs[idx]
-			//console.log("delete : ",idx)
-			if ((b !== null) || (b.visible !== undefined)) {
-				b.visible = false
-			}
-			glowObjs[idx] = null
-		} else if (obj === 'heartbeat') {
-			//console.log("heartbeat")
-		} else if (obj === 'debug') {
-			console.log("debug : ", cmd)
-		}
-		*/
 	} // end of cmds (constructors and special data)
 }
 
-function handle_methods(dmeth) {
+async function handle_methods(dmeth) {
     "use strict";
 	//console.log('METHODS')
-	for (var idmeth=0; idmeth<dmeth.length; idmeth++) { // methods; cmd is ['idx':idx, 'attr':method, 'val':val]
-		var cmd = dmeth[idmeth]
-		var idx = cmd.idx
-		var method = cmd.attr
-		var val = cmd.val
-		var obj = glowObjs[idx]
+	for (let idmeth=0; idmeth<dmeth.length; idmeth++) { // methods; cmd is ['idx':idx, 'attr':method, 'val':val]
+		let cmd = dmeth[idmeth]
+		let idx = cmd.idx
+		let method = cmd.attr
+		let val = cmd.val
+		let obj = glowObjs[idx]
 
 		if (method == 'GSprint') {
 			GSprint(val)
@@ -917,28 +898,29 @@ function handle_methods(dmeth) {
 			glowObjs[idx]['pos'] = val
 		} else if (method === 'add_to_trail') {
 			obj['_func'] = val
-		} else if (method === '_attach_arrow') {
-            obj.obj._attach_arrow = val
 		} else if (method === 'bind') {
-			var evts = val.split(' ')
-			for (var evt in evts) {
-				var e = evts[evt]
+			let evts = val.split(' ')
+			for (let evt in evts) {
+				let e = evts[evt]
 				if (binds.indexOf(e) == -1) 
 					throw new Error('There is no error type "'+e+'"')
 			}
 			obj.bind(val, process_binding)
 		} else if (method === 'unbind') {
-			var evts = val.split(' ')
-			for (var evt in evts) {
-				var e = evts[evt]
+			let evts = val.split(' ')
+			for (let evt in evts) {
+				let e = evts[evt]
 				if (binds.indexOf(e) == -1) 
 					throw new Error('There is no error type "'+e+'"')
 			}
 			obj.unbind(val, process_binding)
 		} else if (method === "follow") {
-			obj.camera.follow(glowObjs[val])
+            if (val === null) obj.camera.follow(null)
+			else obj.camera.follow(glowObjs[val])
 		} else if (method === "capture") {
-			obj.capture(val)
+			// val has the form "Tname.png" (display labels) or "Fname.png" (do not display labels)
+			let TF =  (val[0] == 'T') ? true: false
+			await obj.capture(val.slice(1), TF)
 		} else if (method === 'waitfor') {
 			waitfor_canvas = idx
 			waitfor_options = val
@@ -947,12 +929,13 @@ function handle_methods(dmeth) {
 			waitfor_canvas = idx
 			waitfor_options = 'click'
 			if (val.length > 0) {
-			   obj.pause(val, process_pause) 
+			   await obj.pause(val)
 			} else {
-			   obj.pause(process_pause) 
+			   await obj.pause()
 			}
+			process_pause()
 		} else if (method === 'pick') {
-			var p = glowObjs[val].mouse.pick()   // wait for pick render; val is canvas
+			var p = glowObjs[val].mouse.pick()  // wait for pick render; val is canvas
 			var seg = null
 			if (p !== null) {
 				if (p instanceof curve) seg = p.segment
@@ -966,32 +949,20 @@ function handle_methods(dmeth) {
 function handle_attrs(dattrs) {
     "use strict";
 	//console.log('ATTRS')
-	for (var idattrs=0; idattrs<dattrs.length; idattrs++) { // attributes; cmd is {'idx':idx, 'attr':attr, 'val':val}
-		var cmd = dattrs[idattrs]
-		var idx = cmd.idx
-		var obj = glowObjs[idx]
-		var attr = cmd['attr']
-		var val = cmd['val']
-		var triangle_quad = ['v0', 'v1', 'v2', 'v3']
-		// vector attrs in attach_arrow have arbitrary names, so check for length 3 array instead
+	for (let idattrs=0; idattrs<dattrs.length; idattrs++) { // attributes; cmd is {'idx':idx, 'attr':attr, 'val':val}
+		let cmd = dattrs[idattrs]
+		let idx = cmd.idx
+		let obj = glowObjs[idx]
+		let attr = cmd['attr']
+		let val = cmd['val']
+		let triangle_quad = ['v0', 'v1', 'v2', 'v3']
 		if (val instanceof vec) {
 			if (attr === 'pos' && (obj instanceof points || obj instanceof curve)) {
-				var ptlist = []
-				for (var kk = 0; kk < val.length; kk++) {
+				let ptlist = []
+				for (let kk = 0; kk < val.length; kk++) {
 					ptlist.push( val[kk] )
 				}
 				obj[attr] = ptlist
-			} else if (attr === 'axis') {
-				// For axis and up, Python maintains them to be perpendicular, so avoid
-				//   having GlowScript do it again, which can have bad results.
-				window.__adjustupaxis = false
-				if (obj instanceof arrow) attr = 'axis_and_length'
-				obj[attr] = val
-				window.__adjustupaxis = true
-			} else if (attr === 'up') {
-				window.__adjustupaxis = false
-				obj[attr] = val
-				window.__adjustupaxis = true
 			} else {
 				obj[attr] = val
 			}
